@@ -1,6 +1,6 @@
 GRM <- function(x, chunk = 1L) {
-  if (!x@standardize_mu_sigma & !x@standardize_p) {
-    if (!is.null(x@mu) & !is.null(x@sigma)) x@standardize_p <- TRUE
+  if(!x@standardize_mu_sigma & !x@standardize_p) {
+    if(!is.null(x@p)) x@standardize_p <- TRUE
     else stop("Can't center/scale x for LD computation (use set.stat)\n")
   }
 
@@ -19,3 +19,18 @@ GRM <- function(x, chunk = 1L) {
 
   K
 }
+
+reshape.GRM <- function(K, include = c(-Inf, +Inf), exclude) {
+  diag(K) <- NA
+  if(missing(exclude))
+    w <- which(include[1] < K & K < include[2])
+  else 
+    w <- which(include[1] < K & K < include[2] & (K < exclude[1] | K > exclude[2]))
+  I <- row(K)[w]
+  J <- col(K)[w]
+  R <- K[w]
+  ww <- (I < J)
+  i <- I[ww];
+  j <- J[ww];
+  data.frame(i = i, j = j, id_i = rownames(K)[i], id_j = colnames(K)[j], k = R[ww])
+}
