@@ -12,6 +12,7 @@ void AIREML1(const Eigen::MatrixBase<T1> & y, const Eigen::MatrixBase<T3> & x, c
               double eps, bool verbose, Vector2d & theta, double & logL, double & logL0, int & niter, 
               double & gr_norm, VectorXd & Py, VectorXd & KPy, VectorXd & beta, MatrixXd & XViX_i, double & varXbeta, bool start_theta) {
 
+  Rcout << "aireml1 (fix)\n";
   int n(y.rows()), p(x.cols());
  
   MatrixXd V(n,n);
@@ -33,7 +34,7 @@ void AIREML1(const Eigen::MatrixBase<T1> & y, const Eigen::MatrixBase<T3> & x, c
   double s2_0 = (y.col(0).dot(y.col(0)) - xty.dot( xtxi*xty ))/(n-p);
   logL0 = -0.5*((n-p)*log(s2_0) + ldet_xtx + (n-p));
  
-  Vector2d theta0, gr;
+  Vector2d theta0, gr, gr_cst;
   Matrix2d AI;
   double log_detV, detV, old_logL, d, log_d;
 
@@ -81,7 +82,7 @@ void AIREML1(const Eigen::MatrixBase<T1> & y, const Eigen::MatrixBase<T3> & x, c
         EMsteps = EMsteps_fail;
         if(verbose) Rcout << "[Iteration " << i+1 << "] AI algorithm failed to improve likelihood, trying " 
                           << EMsteps << "EM step\n";
-        theta = theta0;
+        // theta = theta0;
         continue;
       }
     }
@@ -105,11 +106,34 @@ void AIREML1(const Eigen::MatrixBase<T1> & y, const Eigen::MatrixBase<T3> & x, c
       EMsteps--;
     } else {
 
-      if(constraint) { // on débloque les paramètres si le gradient ne pointe plus hors de la boîte
-        if(bloc_s2  && gr(0) > 0) bloc_s2  = false; 
-        if(bloc_tau && gr(1) > 0) bloc_tau = false;
+      if(constraint) {
+        // gradient contraint 
+        if(bloc_s2)  gr_cst(0) = 0; else gr_cst(0) = gr(0);
+        if(bloc_tau) gr_cst(1) = 0; else gr_cst(1) = gr(1);
+
+        // si on a convergé avec la contrainte, on regarde s'il faut débloquer des paramètres 
+        // ie si le gradient ne pointe plus hors de la boîte
+        if( gr_cst.norm() < eps && (bloc_s2 || bloc_tau) ) {
+          if(verbose) Rcout << "[Iteration " << i+1 << "] Checking gradient components signs before last iteration\n";
+          if(bloc_s2) { 
+             if(gr(0) > 0) { 
+               bloc_s2  = false;  
+               gr_cst(0) = gr(0); 
+               if(verbose) Rcout << "  Releasing constraint on sigma^2\n";
+             }
+          }
+          if(bloc_tau) {
+             if(gr(1) > 0) { 
+               bloc_tau = false; 
+               gr_cst(1) = gr(1); 
+               if(verbose) Rcout << "  Releasing constraint on tau\n";
+             }
+          }
+        }
+        gr = gr_cst;
       }
 
+// Rcout << "gradient projeté = " << gr.transpose() << "\n";
       // Average Information
       AI(0,0) = 0.5*PPy.dot(Py);
       AI(1,0) = AI(0,1) = 0.5*PPy.dot(KPy);
@@ -119,21 +143,41 @@ void AIREML1(const Eigen::MatrixBase<T1> & y, const Eigen::MatrixBase<T3> & x, c
 
       if(constraint && bloc_s2) {
         theta(1) += gr(1)/AI(1,1);
-        gr(0) = 0;
       }
       else if(constraint && bloc_tau) {
         theta(0) += gr(0)/AI(0,0);
-        gr(1) = 0;
       }
       else {
         theta += AI.inverse()*gr;
       }
-
+// Rcout << "theta = " << theta.transpose() << "\n";
       if(constraint) {
-        if(theta(0)  < min_s2)  { theta(0) = min_s2;  bloc_s2  = true; }
-        if(theta(1)  < min_tau) { theta(1) = min_tau; bloc_tau = true; }
+        double lambda = 1;
+        int a_bloquer = -1; // nécessaire pour pallier aux erreurs d'arrondi
+        if(theta(0) < min_s2) {
+           double lambda0 = (min_s2 - theta0(0))/(theta(0)-theta0(0));
+           if(lambda0 < lambda) { // forcément vrai ici...
+              lambda = lambda0;
+              a_bloquer = 0;
+           }
+        }
+        if(theta(1) < min_tau) {
+           double lambda0 = (min_tau - theta0(1))/(theta(1)-theta0(1));
+           if(lambda0 < lambda) { // ...mais pas ici !
+              lambda = lambda0;
+              a_bloquer = 1;
+           }
+        }
+        theta = theta0 + lambda*(theta-theta0);
+        if(theta(0) < min_s2  || a_bloquer == 0) { 
+          theta(0) = min_s2;  bloc_s2  = true; 
+          if(verbose) Rcout << "[Iteration " << i+1 << "] Constraining sigma^2\n";
+        }
+        if(theta(1) < min_tau || a_bloquer == 1) { 
+          theta(1) = min_tau; bloc_tau = true; 
+          if(verbose) Rcout << "[Iteration " << i+1 << "] Constraining tau\n";
+        }
       }
-
       if(verbose) Rcout << "[Iteration " << i+1 << "] AI-REML update" << "\n";
       EM = false;
     }
